#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
/* #include <sys/socket.h> */
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <termios.h>
#include <unistd.h>
#include <utmp.h>
#include <values.h>
#include <time.h>
#include <linux/socket.h>
/* #include <linux/in.h> */

#define TIMEOUT 300

extern char *Version;

/*
 * Set the terminal options we will use
 */
void set_term_opt(int term)
{
	struct termios trm;

	tcgetattr(term, &trm);
	trm.c_lflag &= ~ICANON;
	trm.c_lflag &= ~ECHO;
	trm.c_iflag &= ~IGNCR;
	trm.c_oflag &= ~ONLRET;
	trm.c_oflag &= ~ONOCR;
	tcsetattr(term, TCSANOW, &trm);
}

/*
 * Allocate a pty.
 * Returns the file descriptor of the opened pty and put the name 
 * of the slave side in line.
 */
int get_pty(char *line)
{
	register int p;
	register char c, *p1, *p2;
	register int i;

	sprintf(line, "/dev/ptyXX");
	p1 = &line[8];
	p2 = &line[9];

	for (c = 'p'; c <= 's'; c++) {
		struct stat stb;

		*p1 = c;
		*p2 = '0';
		if (stat(line, &stb) < 0)
			break;
		for (i = 0; i < 16; i++) {
			*p2 = "0123456789abcdef"[i];
			p = open(line, O_RDWR);
			if (p > 0) {
				line[5] = 't';
				return (p);
			}
		}
	}
	return -1;
}

/*
 * Close the ftmp file setting up appropriate ownership
 */
void close_ftmp(int f, char *user)
{
	struct passwd *pwd;
	int res;

	if ((pwd = getpwnam(user)) != NULL) {
		res = fchown(f, pwd->pw_uid, pwd->pw_gid);
	}
	close(f);
}

/*
 * Returns date&time in the DDMmmHHMMSS format
 */
char *mydate(void)
{
	time_t t;
	struct tm *tim;
	static char buf[80];

	time(&t);
	tim = gmtime(&t);
	strftime(buf, 80, "%d%b%H%M%S", tim);
	return buf;
}

/*
 * Notify sysop of a message
 */
void mailsysop(int fno, char *user, char *mailto)
{
	char buf[128];
	FILE *fp;
	int len;

	if (fno == 0)
		return;

	sprintf(buf, "mail -s \"TTYlinkd-saved message from %s\" %s", user, mailto);
	if ((fp=popen(buf,"w")) != NULL) {
		lseek(fno, 0, SEEK_SET);
		while ((len=read(fno, buf, 128)) > 0)
			fwrite(buf, 1, len, fp);
		pclose(fp);
	}
}	

/*
 * Returns the minimum idletime for a tty
 */
time_t idletime(char *tty)
{
	struct stat st;
	char ttyname[20];
	time_t t, at, mt, ct;

	sprintf(ttyname,"/dev/%s", tty);

	if (stat(ttyname, &st) != 0)
		return 0;
	
	t = time(NULL);
	at = t - st.st_atime;
	ct = t - st.st_ctime;
	mt = t - st.st_mtime;
#ifdef DEBUG
	printf("Idletimes: at=%d ct=%d mt=%d\n", at, ct, mt);
#endif
	return (at < ct) ? (mt < at) ? mt : at : (mt < ct) ? mt : ct; 
}

/*
 * Search if one of the defined users is active.
 * Return 1 if one is found, 0 otherwise.
 * If found, copy in up the utmp record of the tty most recently used.
 */
int search_user(char *users, struct utmp *up)
{
	struct utmp *utp;
	int found = 0;
	time_t minidle = MAXINT, idle;

	setutent();
	while ((utp = getutent()) != NULL) {
#ifdef DEBUG
		printf("Found user: %s Line: %s Host: %s Type: %d\n",utp->ut_user,utp->ut_line,utp->ut_host,utp->ut_type);
#endif
		if ((utp->ut_type == USER_PROCESS) &&
		    (*utp->ut_user) &&
		    (*utp->ut_line != ':') &&
		    (strstr(users, utp->ut_user) != NULL)) {
		  	if ((idle=idletime(utp->ut_line)) < minidle) {
#ifdef DEBUG
				printf("Returned idletime = %d\n", idle);
#endif
				memcpy(up, utp, sizeof(struct utmp));
				minidle = idle;
				found = 1;
			}
		}
	}
	endutent();
	return found;
}

int main(int argc, char *argv[])
{
	int f1, f2, f3, res, len, found, no_response, have_data;
	struct sockaddr_in sock_in;
	fd_set fset;
	char buf[1024], caller[256], users[256], *cp;
	struct hostent *hname;
	struct utmp ut;
	struct timeval tval;
	char ttyname[80], ptyname[80], filename[80];

	signal(SIGCLD, SIG_IGN);

	len = sizeof(sock_in);
	if ((getpeername(0, (struct sockaddr *) &sock_in, &len)) == -1) {
		perror("getpeername");
		strcpy(caller, "user");
	} else {
		hname = gethostbyaddr((char *) &sock_in.sin_addr.s_addr, sizeof(sock_in.sin_addr.s_addr), AF_INET);
		strcpy(caller, hname->h_name);
	}

	printf("\r\nWelcome %s to %s !\n\n", caller, Version);
	fflush(stdout);

	found = 0;
	if (argc > 1)
		strcpy(users, argv[1]);
	else
		strcpy(users,"root");

	found = search_user(users, &ut);

	if (!found) {
		printf("None of [%s] is currently logged in.\n", users);
#ifdef DISCONNECT_IF_NONE
		printf("Try again later.\n");
		fflush(stdout);
		exit(0);
#else
		cp = strtok(users,",");
		strcpy(ut.ut_user, cp);
		printf("If you write something, it will be mailed to %s.\n",cp);
		fflush(stdout);
#endif
	} else {
		printf("%s is logged in, please wait, ringing bells...\n\n", ut.ut_user);
		fflush(stdout);
	}

	if (found) {
		sprintf(ttyname, "/dev/%s", ut.ut_line);
#ifdef DEBUG
		printf("Trying to open %s\n",ttyname);
#endif
		if ((f1 = open(ttyname, O_RDWR)) == -1) {
			perror("open()");
			exit(1);
		}
	
		if ((f2 = get_pty(ptyname)) == -1) {
			perror("get_pty()");
			exit(1);
		}
	
		set_term_opt(f2);
	}
	if (strcmp(ut.ut_user, "root") == 0) {
		sprintf(filename, "/root/TTYLINK-%s-%s", caller, mydate());
	} else {
		sprintf(filename, "/home/%s/TTYLINK-%s-%s", ut.ut_user, caller, mydate());
	}

	if ((f3 = open(filename, O_RDWR | O_CREAT, 0x1b6)) == -1) {
		perror("open()");
		exit(1);
	}
	if (found) {
		sprintf(buf, "\n\n\r\a\aMessage from TTYlink daemon:\r\n\tIncoming call from %s, type ttydriv %s to answer.\a\a\r\n\n", caller, ptyname);
		res = write(f1, buf, strlen(buf));
	}

	no_response = 1;
	have_data = 0;
	while (1) {
		FD_ZERO(&fset);
		FD_SET(STDIN_FILENO, &fset);
		if (found)
			FD_SET(f2, &fset);
		tval.tv_sec = TIMEOUT;
		tval.tv_usec = 0;

		if (found)
			res = select(f2 + 1, &fset, NULL, NULL, &tval);
		else
			res = select(STDIN_FILENO + 1, &fset, NULL, NULL, &tval);
		
		fflush(stdout);

		if ((res == 0) && (no_response)) {
			if (found)
				printf("\nNo response from %s in %d minutes. Disconnecting.\n", ut.ut_user, TIMEOUT/60);
			else
				printf("\nNo-data %d minutes timeout elapsed. Disconnecting.\n", TIMEOUT/60);
			if (have_data) {
				printf("Received data have been mailed to %s and saved in the file:\n\t%s\n", ut.ut_user, filename);
			}
			fflush(stdout);
			cp=buf;
			cp += sprintf(cp, "\r\n\nMessage from TTYlink daemon:\r\n\tTimeout, %s has been disconnected.\r\n", caller);
			if (found) {
				if (have_data) {
					mailsysop(f3, caller, ut.ut_user); 
					cp += sprintf(cp,"\tReceived data from %s have been mailed to %s\r\n\tand stored in the file: %s\r\n\n", caller, ut.ut_user, filename);
				} else {
					cp += sprintf(cp, "\r\n");
				}
				res = write(f1, buf, strlen(buf));
			} else if (have_data) {
				mailsysop(f3, caller, ut.ut_user); 
			}
			close_ftmp(f3, ut.ut_user);
			if (!have_data)
				unlink(filename);
			close(0);
			exit(0);
		}
		if (res == -1) {
			perror("select()");
			exit(2);
		}
		if (FD_ISSET(0, &fset)) {
			len = read(0, buf, sizeof(buf));
			if (len <= 0) {
				if (found) {
					cp=buf;
					cp += sprintf(cp, "\r\n\nMessage from TTYlink daemon:\r\n\t%s has closed connection.\r\n", caller);
					if (have_data && no_response) {
						mailsysop(f3, caller, ut.ut_user); 
						cp += sprintf(cp,"\r\tReceived data from %s have been mailed to %s\r\n\tand stored in the file: %s\n\n", caller, ut.ut_user, filename);
					} else {
						cp += sprintf(cp, "\n");
					}
					res = write(f1, buf, strlen(buf));
				} else if (have_data && no_response) {
					mailsysop(f3, caller, ut.ut_user);
				}
				close_ftmp(f3, ut.ut_user);
				if (!have_data)
					unlink(filename);
				exit(2);
			}
			if (no_response) {
				res = write(f3, buf, len);
			} else {
				res = write(f2, buf, len);
			}
			have_data = 1;
			continue;
		}
		if (FD_ISSET(f2, &fset)) {
			len = read(f2, buf, sizeof(buf));
			if (len <= 0) {
				close_ftmp(f3, ut.ut_user);
				exit(3);
			}
			res = write(STDOUT_FILENO, buf, len);
			fflush(stdout);
			if (no_response) {
				no_response = 0;
				if (found)
					close(f1);
				lseek(f3, 0, SEEK_SET);
				while ((len = read(f3, buf, sizeof(buf))) > 0) {
					res = write(f2, buf, len);
				}
				close(f3);
				unlink(filename);
			}
			continue;
		}
	}
}
